;;;; stack.scm
;;;; Kon Lovett, Mar 29 2005

;;;; Stack data structure (LIFO queue) where the value is mutable,
;;;; rather than usual pattern of the variable.

;; Helpers

(define-macro (s-u$stack-pusher S X)
	`(set-cdr! ,S (cons ,X (cdr ,S))) )

;;;

;; CLEAR-STACK
;;
;; An empty stack.

(define-macro (clear-stack S)
	`(set-cdr ,S '()) )

;; MAKE-STACK
;;
;; Returns an empty stack.

(define-macro (make-stack . ARGS)
	`(cons 'stack ,(if (null? ARGS) ''() (car ARGS))) )

;; STACK? Stack
;;
;; Stack - a stack
;;
;; Returns #t for a stack, #f otherwise. Not fool-proof!

(define-macro (stack? S)
	`(let ([stk ,S])
		(and (pair? stk) (eq? 'stack (car stk)))) )

;; STACK-EMPTY? Stack
;;
;; Stack - a stack
;;
;; Returns #t for an empty stack, #f otherwise.

(define-macro (stack-empty? S)
	`(null? (cdr ,S)) )

;; STACK-LENGTH Stack
;;
;; Stack - a stack
;;
;; Returns the number of elements on the stack.

(define-macro (stack-length S)
	`(length (cdr ,S)) )

;; STACK-DEPTH Stack
;;
;; Stack - a stack
;;
;; Returns the index of the last element on the stack, or -1 when stack
;; empty.

(define-macro (stack-depth S)
	`(fx- (stack-length ,S) 1) )

;; STACK-PEEK Stack #!optional Index
;;
;; Stack - the stack to peek into
;;
;; Index - where to peek, 0 <= & <= (stack-depth
;; Stack) - 1, defaults to 0
;;
;; Returns the peeked value.

(define-macro (stack-peek S . ARGS)
	`(car (s-u$node-ref 'stack-peek ,S ,(if (null? ARGS) 0 (car ARGS)))) )

;; STACK-POKE! Stack Value #!optional Index
;;
;; Stack - the stack to poke into
;;
;; Value - the value to poke
;;
;; Index - where to poke, 0 <= & <= (stack-depth
;; Stack) - 1, defaults to 0
;;
;; Returns the modified stack. The stack is modified in place.

(define-macro (stack-poke! S X . ARGS)
	`(let ([stk ,S])
		(set-car! (s-u$node-ref 'stack-poke! stk ,(if (null? ARGS) 0 (car ARGS))) ,X)
		stk ) )

;; STACK-PUSH! Stack Value ...
;;
;; Stack - the stack to push onto
;;
;; Value - the new top of stack value(s)
;;
;; Returns the modified stack. The stack is modified in place.

(define-macro (stack-push! S . VS)
	(cond [(null? VS)
          S]
        [(null? (cdr VS))
          `(let ([stk ,S])
            (s-u$stack-pusher stk ,(car VS))
            stk )]
        [else
          `(let ([stk ,S])
            (for-each (lambda (x) (s-u$stack-pusher stk x)) (list ,@VS))
            stk )]) )

;; STACK-POP*! Stack
;;
;; Stack - the stack to pop
;;
;; Returns the top of the stack. The stack is modified in place.

(define-macro (stack-pop*! S)
	`(let* ([stk ,S]
				  [x (cadr stk)])
		(set-cdr! stk (cddr stk))
		x) )

;; LIST->STACK List
;;
;; List - a proper list
;;
;; Returns the list as a stack. The resulting stack may share memory
;; with the list and the list should not be modified after this operation.

(define-macro (list->stack L)
	`(make-stack ,L) )

;; STACK->LIST Stack
;;
;; Stack - a stack
;;
;; Returns the stack as a list, where the first element of the
;; list is the top element of the stack.  The resulting list may share
;; memory with the queue object and should not be modified.

(define-macro (stack->list S)
	`(cdr ,S) )

;; STACK-FOR-EACH Stack Procedure
;;
;; Stack - a stack
;;
;; Procedure - a procedure of one argument
;;
;; Applies the procedure to each element of the stack, top to bottom.

(define-macro (stack-for-each S F)
	`(for-each ,F (cdr ,S)) )

;; PRINT-STACK Stack
;;
;; Stack - a stack
;;
;; Prints the elements of the stack.

(define-macro (print-stack S)
	`(stack-for-each ,S print) )
